# -*- coding: utf-8 -*-
from unittest import TestCase

import hl7
from hl7 import Field, Segment

from .samples import sample_hl7


class ContainerTest(TestCase):
    def test_unicode(self):
        msg = hl7.parse(sample_hl7)
        self.assertEqual(str(msg), sample_hl7)
        self.assertEqual(
            str(msg[3][3]), "1554-5^GLUCOSE^POST 12H CFST:MCNC:PT:SER/PLAS:QN"
        )

    def test_container_unicode(self):
        c = hl7.Container("|")
        c.extend(["1", "b", "data"])
        self.assertEqual(str(c), "1|b|data")


class MessageTest(TestCase):
    def test_segments(self):
        msg = hl7.parse(sample_hl7)
        s = msg.segments("OBX")
        self.assertEqual(len(s), 2)
        self.assertIsInstance(s[0], Segment)
        self.assertEqual(s[0][0:3], [["OBX"], ["1"], ["SN"]])
        self.assertEqual(s[1][0:3], [["OBX"], ["2"], ["FN"]])

        self.assertIsInstance(s[0][1], Field)

    def test_segments_does_not_exist(self):
        msg = hl7.parse(sample_hl7)
        self.assertRaises(KeyError, msg.segments, "BAD")

    def test_segment(self):
        msg = hl7.parse(sample_hl7)
        s = msg.segment("OBX")
        self.assertEqual(s[0:3], [["OBX"], ["1"], ["SN"]])

    def test_segment_does_not_exist(self):
        msg = hl7.parse(sample_hl7)
        self.assertRaises(KeyError, msg.segment, "BAD")

    def test_segments_dict_key(self):
        msg = hl7.parse(sample_hl7)
        s = msg["OBX"]
        self.assertEqual(len(s), 2)
        self.assertEqual(s[0][0:3], [["OBX"], ["1"], ["SN"]])
        self.assertEqual(s[1][0:3], [["OBX"], ["2"], ["FN"]])

    def test_MSH_1_field(self):
        msg = hl7.parse(sample_hl7)
        f = msg["MSH.1"]
        self.assertEqual(len(f), 1)
        self.assertEqual(f, "|")

    def test_MSH_2_field(self):
        msg = hl7.parse(sample_hl7)
        f = msg["MSH.2"]
        self.assertEqual(len(f), 4)
        self.assertEqual(f, "^~\\&")

    def test_get_slice(self):
        msg = hl7.parse(sample_hl7)
        s = msg.segments("OBX")[0]
        self.assertIsInstance(s, Segment)
        self.assertIsInstance(s[0:3], Segment)

    def test_ack(self):
        msg = hl7.parse(sample_hl7)
        ack = msg.create_ack()
        self.assertEqual(msg["MSH.1"], ack["MSH.1"])
        self.assertEqual(msg["MSH.2"], ack["MSH.2"])
        self.assertEqual("ACK", ack["MSH.9.1.1"])
        self.assertEqual(msg["MSH.9.1.2"], ack["MSH.9.1.2"])
        self.assertEqual("ACK", ack["MSH.9.1.3"])
        self.assertNotEqual(msg["MSH.7"], ack["MSH.7"])
        self.assertNotEqual(msg["MSH.10"], ack["MSH.10"])
        self.assertEqual("AA", ack["MSA.1"])
        self.assertEqual(msg["MSH.10"], ack["MSA.2"])
        self.assertEqual(20, len(ack["MSH.10"]))
        self.assertEqual(msg["MSH.5"], ack["MSH.3"])
        self.assertEqual(msg["MSH.6"], ack["MSH.4"])
        self.assertEqual(msg["MSH.3"], ack["MSH.5"])
        self.assertEqual(msg["MSH.4"], ack["MSH.6"])
        ack2 = msg.create_ack(
            ack_code="AE", message_id="testid", application="python", facility="test"
        )
        self.assertEqual("AE", ack2["MSA.1"])
        self.assertEqual("testid", ack2["MSH.10"])
        self.assertEqual("python", ack2["MSH.3"])
        self.assertEqual("test", ack2["MSH.4"])
        self.assertNotEqual(ack["MSH.10"], ack2["MSH.10"])


class TestMessage(hl7.Message):
    pass


class TestSegment(hl7.Segment):
    pass


class TestField(hl7.Field):
    pass


class TestRepetition(hl7.Repetition):
    pass


class TestComponent(hl7.Component):
    pass


class TestFactory(hl7.Factory):
    create_message = TestMessage
    create_segment = TestSegment
    create_field = TestField
    create_repetition = TestRepetition
    create_component = TestComponent


class FactoryTest(TestCase):
    def test_parse(self):
        msg = hl7.parse(sample_hl7, factory=TestFactory)
        self.assertIsInstance(msg, TestMessage)
        s = msg.segments("OBX")
        self.assertIsInstance(s[0], TestSegment)
        self.assertIsInstance(s[0](3), TestField)
        self.assertIsInstance(s[0](3)(1), TestRepetition)
        self.assertIsInstance(s[0](3)(1)(1), TestComponent)
        self.assertEqual("1554-5", s[0](3)(1)(1)(1))

    def test_ack(self):
        msg = hl7.parse(sample_hl7, factory=TestFactory)
        ack = msg.create_ack()
        self.assertIsInstance(ack, TestMessage)
        self.assertIsInstance(ack(1)(9), TestField)
        self.assertIsInstance(ack(1)(9)(1), TestRepetition)
        self.assertIsInstance(ack(1)(9)(1)(2), TestComponent)
        self.assertEqual("R01", ack(1)(9)(1)(2)(1))



# -*- coding: utf-8 -*-
from collections import namedtuple


class Accessor(
    namedtuple(
        "Accessor",
        [
            "segment",
            "segment_num",
            "field_num",
            "repeat_num",
            "component_num",
            "subcomponent_num",
        ],
    )
):
    __slots__ = ()

    def __new__(
        cls,
        segment,
        segment_num=1,
        field_num=None,
        repeat_num=None,
        component_num=None,
        subcomponent_num=None,
    ):
        """Create a new instance of Accessor for *segment*. Index numbers start from 1."""
        return super(Accessor, cls).__new__(
            cls,
            segment,
            segment_num,
            field_num,
            repeat_num,
            component_num,
            subcomponent_num,
        )

    @property
    def key(self):
        """Return the string accessor key that represents this instance"""
        seg = (
            self.segment
            if self.segment_num == 1
            else self.segment + str(self.segment_num)
        )
        return ".".join(
            str(f)
            for f in [
                seg,
                self.field_num,
                self.repeat_num,
                self.component_num,
                self.subcomponent_num,
            ]
            if f is not None
        )

    def __str__(self):
        return self.key

    @classmethod
    def parse_key(cls, key):
        """Create an Accessor by parsing an accessor key.

        The key is defined as:

            |   SEG[n]-Fn-Rn-Cn-Sn
            |       F   Field
            |       R   Repeat
            |       C   Component
            |       S   Sub-Component
            |
            |   *Indexing is from 1 for compatibility with HL7 spec numbering.*

        Example:

            |   PID.F1.R1.C2.S2 or PID.1.1.2.2
            |
            |   PID (default to first PID segment, counting from 1)
            |   F1  (first after segment id, HL7 Spec numbering)
            |   R1  (repeat counting from 1)
            |   C2  (component 2 counting from 1)
            |   S2  (component 2 counting from 1)
        """

        def parse_part(keyparts, index, prefix):
            if len(keyparts) > index:
                num = keyparts[index]
                if num[0].upper() == prefix:
                    num = num[1:]
                return int(num)
            else:
                return None

        parts = key.split(".")
        segment = parts[0][:3]
        if len(parts[0]) > 3:
            segment_num = int(parts[0][3:])
        else:
            segment_num = 1
        field_num = parse_part(parts, 1, "F")
        repeat_num = parse_part(parts, 2, "R")
        component_num = parse_part(parts, 3, "C")
        subcomponent_num = parse_part(parts, 4, "S")
        return cls(
            segment, segment_num, field_num, repeat_num, component_num, subcomponent_num
        )



import io
import os.path
import socket
import sys
from optparse import OptionParser

import hl7

SB = b"\x0b"  # <SB>, vertical tab
EB = b"\x1c"  # <EB>, file separator
CR = b"\x0d"  # <CR>, \r

FF = b"\x0c"  # <FF>, new page form feed

RECV_BUFFER = 4096


class MLLPException(Exception):
    pass


class MLLPClient(object):
    """
    A basic, blocking, HL7 MLLP client based upon :py:mod:`socket`.

    MLLPClient implements two methods for sending data to the server.

    * :py:meth:`MLLPClient.send` for raw data that already is wrapped in the
      appropriate MLLP container (e.g. *<SB>message<EB><CR>*).
    * :py:meth:`MLLPClient.send_message` will wrap the message in the MLLP
      container

    Can be used by the ``with`` statement to ensure :py:meth:`MLLPClient.close`
    is called::

        with MLLPClient(host, port) as client:
            client.send_message('MSH|...')

    MLLPClient takes an optional ``encoding`` parameter, defaults to UTF-8,
    for encoding unicode messages [#]_.

    .. [#] http://wiki.hl7.org/index.php?title=Character_Set_used_in_v2_messages
    """

    def __init__(self, host, port, encoding="utf-8"):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((host, port))
        self.encoding = encoding

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, trackeback):
        self.close()

    def close(self):
        """Release the socket connection"""
        self.socket.close()

    def send_message(self, message):
        """Wraps a byte string, unicode string, or :py:class:`hl7.Message`
        in a MLLP container and send the message to the server

        If message is a byte string, we assume it is already encoded properly.
        If message is unicode or  :py:class:`hl7.Message`, it will be encoded
        according to  :py:attr:`hl7.client.MLLPClient.encoding`

        """
        if isinstance(message, bytes):
            # Assume we have the correct encoding
            binary = message
        else:
            # Encode the unicode message into a bytestring
            if isinstance(message, hl7.Message):
                message = str(message)
            binary = message.encode(self.encoding)

        # wrap in MLLP message container
        data = SB + binary + EB + CR
        return self.send(data)

    def send(self, data):
        """Low-level, direct access to the socket.send (data must be already
        wrapped in an MLLP container).  Blocks until the server returns.
        """
        # upload the data
        self.socket.send(data)
        # wait for the ACK/NACK
        return self.socket.recv(RECV_BUFFER)


# wrappers to make testing easier
def stdout(content):
    # In Python 3, can't write bytes via sys.stdout.write
    #   http://bugs.python.org/issue18512
    if isinstance(content, bytes):
        out = sys.stdout.buffer
        newline = b"\n"
    else:
        out = sys.stdout
        newline = "\n"

    out.write(content + newline)


def stdin():
    return sys.stdin


def stderr():
    return sys.stderr


def read_stream(stream):
    """Buffer the stream and yield individual, stripped messages"""
    _buffer = b""

    while True:
        data = stream.read(RECV_BUFFER)
        if data == b"":
            break
        # usually should be broken up by EB, but I have seen FF separating
        # messages
        messages = (_buffer + data).split(EB if FF not in data else FF)

        # whatever is in the last chunk is an uncompleted message, so put back
        # into the buffer
        _buffer = messages.pop(-1)

        for m in messages:
            yield m.strip(SB + CR)

    if len(_buffer.strip()) > 0:
        raise MLLPException("buffer not terminated: %s" % _buffer)


def read_loose(stream):
    """Turn a HL7-like blob of text into a real HL7 messages"""
    # look for the START_BLOCK to delineate messages
    START_BLOCK = rb"MSH|^~\&|"

    # load all the data
    data = stream.read()

    # Take out all the typical MLLP separators. In Python 3, iterating
    # through a bytestring returns ints, so we need to filter out the int
    # versions of the separators, then convert back from a list of ints to
    # a bytestring.
    # WARNING: There is an assumption here that we can treat the data as single bytes
    #   when filtering out the separators.
    separators = [bs[0] for bs in [EB, FF, SB]]
    data = bytes(b for b in data if b not in separators)
    # Windows & Unix new lines to segment separators
    data = data.replace(b"\r\n", b"\r").replace(b"\n", b"\r")

    for m in data.split(START_BLOCK):
        if not m:
            # the first element will not have any data from the split
            continue

        # strip any trailing whitespace
        m = m.strip(CR + b"\n ")

        # re-insert the START_BLOCK, which was removed via the split
        yield START_BLOCK + m


def mllp_send():
    """Command line tool to send messages to an MLLP server"""
    # set up the command line options
    script_name = os.path.basename(sys.argv[0])
    parser = OptionParser(usage=script_name + " [options] <server>")
    parser.add_option(
        "--version",
        action="store_true",
        dest="version",
        default=False,
        help="print current version and exit",
    )
    parser.add_option(
        "-p",
        "--port",
        action="store",
        type="int",
        dest="port",
        default=6661,
        help="port to connect to",
    )
    parser.add_option(
        "-f",
        "--file",
        dest="filename",
        help="read from FILE instead of stdin",
        metavar="FILE",
    )
    parser.add_option(
        "-q",
        "--quiet",
        action="store_true",
        dest="verbose",
        default=True,
        help="do not print status messages to stdout",
    )
    parser.add_option(
        "--loose",
        action="store_true",
        dest="loose",
        default=False,
        help=(
            "allow file to be a HL7-like object (\\r\\n instead "
            "of \\r). Requires that messages start with "
            '"MSH|^~\\&|". Requires --file option (no stdin)'
        ),
    )

    (options, args) = parser.parse_args()

    if options.version:
        import hl7

        stdout(hl7.__version__)
        return

    if len(args) == 1:
        host = args[0]
    else:
        # server not present
        parser.print_usage()
        stderr().write("server required\n")
        sys.exit(1)
        return  # for testing when sys.exit mocked

    if options.filename is not None:
        # Previously set stream to the open() handle, but then we did not
        # close the open file handle.  This new approach consumes the entire
        # file into memory before starting to process, which is not required
        # or ideal, since we can handle a stream
        with open(options.filename, "rb") as f:
            stream = io.BytesIO(f.read())
    else:
        if options.loose:
            stderr().write("--loose requires --file\n")
            sys.exit(1)
            return  # for testing when sys.exit mocked

        stream = stdin()

    with MLLPClient(host, options.port) as client:
        message_stream = (
            read_stream(stream) if not options.loose else read_loose(stream)
        )

        for message in message_stream:
            result = client.send_message(message)
            if options.verbose:
                stdout(result)


if __name__ == "__main__":
    mllp_send()



import warnings
from asyncio import (
    LimitOverrunError,
    StreamReader,
    StreamReaderProtocol,
    StreamWriter,
    get_event_loop,
    iscoroutine,
)
from asyncio.streams import _DEFAULT_LIMIT

from hl7.mllp.exceptions import InvalidBlockError
from hl7.parser import parse as hl7_parse

START_BLOCK = b"\x0b"
END_BLOCK = b"\x1c"
CARRIAGE_RETURN = b"\x0d"


async def open_hl7_connection(
    host=None,
    port=None,
    *,
    loop=None,
    limit=_DEFAULT_LIMIT,
    encoding=None,
    encoding_errors=None,
    **kwds
):
    """A wrapper for `loop.create_connection()` returning a (reader, writer) pair.

    The reader returned is a :py:class:`hl7.mllp.HL7StreamReader` instance; the writer is a
    :py:class:`hl7.mllp.HL7StreamWriter` instance.

    The arguments are all the usual arguments to create_connection()
    except `protocol_factory`; most common are positional `host` and `port`,
    with various optional keyword arguments following.

    Additional optional keyword arguments are `loop` (to set the event loop
    instance to use), `limit` (to set the buffer limit passed to the
    :py:class:`hl7.mllp.HL7StreamReader`), `encoding` (to set the encoding on the :py:class:`hl7.mllp.HL7StreamReader`
    and :py:class:`hl7.mllp.HL7StreamWriter`) and `encoding_errors` (to set the encoding_errors on the :py:class:`hl7.mllp.HL7StreamReader`
    and :py:class:`hl7.mllp.HL7StreamWriter`).
    """
    if loop is None:
        loop = get_event_loop()
    else:
        warnings.warn(
            "The loop argument is deprecated since Python 3.8, "
            "and scheduled for removal in Python 3.10.",
            DeprecationWarning,
            stacklevel=2,
        )
    reader = HL7StreamReader(
        limit=limit, loop=loop, encoding=encoding, encoding_errors=encoding_errors
    )
    protocol = HL7StreamProtocol(
        reader, loop=loop, encoding=encoding, encoding_errors=encoding_errors
    )
    transport, _ = await loop.create_connection(lambda: protocol, host, port, **kwds)
    writer = HL7StreamWriter(
        transport, protocol, reader, loop, encoding, encoding_errors
    )
    return reader, writer


async def start_hl7_server(
    client_connected_cb,
    host=None,
    port=None,
    *,
    loop=None,
    limit=_DEFAULT_LIMIT,
    encoding=None,
    encoding_errors=None,
    **kwds
):
    """Start a socket server, call back for each client connected.

    The first parameter, `client_connected_cb`, takes two parameters:
    `client_reader`, `client_writer`.  `client_reader` is a
    :py:class:`hl7.mllp.HL7StreamReader` object, while `client_writer`
    is a :py:class:`hl7.mllp.HL7StreamWriter` object.  This
    parameter can either be a plain callback function or a coroutine;
    if it is a coroutine, it will be automatically converted into a
    `Task`.

    The rest of the arguments are all the usual arguments to
    `loop.create_server()` except `protocol_factory`; most common are
    positional `host` and `port`, with various optional keyword arguments
    following.

    The return value is the same as `loop.create_server()`.
    Additional optional keyword arguments are `loop` (to set the event loop
    instance to use) and `limit` (to set the buffer limit passed to the
    StreamReader).

    The return value is the same as `loop.create_server()`, i.e. a
    `Server` object which can be used to stop the service.
    """
    if loop is None:
        loop = get_event_loop()
    else:
        warnings.warn(
            "The loop argument is deprecated since Python 3.8, "
            "and scheduled for removal in Python 3.10.",
            DeprecationWarning,
            stacklevel=2,
        )

    def factory():
        reader = HL7StreamReader(
            limit=limit, loop=loop, encoding=encoding, encoding_errors=encoding_errors
        )
        protocol = HL7StreamProtocol(
            reader,
            client_connected_cb,
            loop=loop,
            encoding=encoding,
            encoding_errors=encoding_errors,
        )
        return protocol

    return await loop.create_server(factory, host, port, **kwds)


class MLLPStreamReader(StreamReader):
    def __init__(self, limit=_DEFAULT_LIMIT, loop=None):
        super().__init__(limit, loop)

    async def readblock(self):
        """Read a chunk of data from the stream until the block termination
        separator (b'\x1c\x0d') are found.

        On success, the data and separator will be removed from the
        internal buffer (consumed). Returned data will not include the
        separator at the end or the MLLP start block character (b'\x0b') at the
        beginning.

        Configured stream limit is used to check result. Limit sets the
        maximal length of data that can be returned, not counting the
        separator.

        If an EOF occurs and the complete separator is still not found,
        an IncompleteReadError exception will be raised, and the internal
        buffer will be reset.  The IncompleteReadError.partial attribute
        may contain the separator partially.

        If limit is reached, ValueError will be raised. In that case, if
        block termination separator was found, complete line including separator
        will be removed from internal buffer. Else, internal buffer will be cleared. Limit is
        compared against part of the line without separator.

        If the block is invalid (missing required start block character) and InvalidBlockError
        will be raised.

        If stream was paused, this function will automatically resume it if
        needed.
        """
        sep = END_BLOCK + CARRIAGE_RETURN
        seplen = len(sep)
        try:
            block = await self.readuntil(sep)
        except LimitOverrunError as loe:
            if self._buffer.startswith(sep, loe.consumed):
                del self._buffer[: loe.consumed + seplen]
            else:
                self._buffer.clear()
            self._maybe_resume_transport()
            raise ValueError(loe.args[0])
        if not block or block[0:1] != START_BLOCK:
            raise InvalidBlockError(
                "Block does not begin with Start Block character <VT>"
            )
        return block[1:-2]


class MLLPStreamWriter(StreamWriter):
    def __init__(self, transport, protocol, reader, loop):
        super().__init__(transport, protocol, reader, loop)

    def writeblock(self, data):
        """Write a block of data to the stream,
        encapsulating the block with b'\x0b' at the beginning
        and b'\x1c\x0d' at the end.
        """
        self.write(START_BLOCK + data + END_BLOCK + CARRIAGE_RETURN)


class HL7StreamProtocol(StreamReaderProtocol):
    def __init__(
        self,
        stream_reader,
        client_connected_cb=None,
        loop=None,
        encoding=None,
        encoding_errors=None,
    ):
        super().__init__(stream_reader, client_connected_cb, loop)
        self._encoding = encoding
        self._encoding_errors = encoding_errors

    def connection_made(self, transport):
        if self._reject_connection:
            context = {
                "message": (
                    "An open stream was garbage collected prior to "
                    "establishing network connection; "
                    'call "stream.close()" explicitly.'
                )
            }
            if self._source_traceback:
                context["source_traceback"] = self._source_traceback
            self._loop.call_exception_handler(context)
            transport.abort()
            return
        self._transport = transport
        reader = self._stream_reader
        if reader is not None:
            reader.set_transport(transport)
        self._over_ssl = transport.get_extra_info("sslcontext") is not None
        if self._client_connected_cb is not None:
            self._stream_writer = HL7StreamWriter(
                transport,
                self,
                reader,
                self._loop,
                self._encoding,
                self._encoding_errors,
            )
            res = self._client_connected_cb(reader, self._stream_writer)
            if iscoroutine(res):
                self._loop.create_task(res)
            self._strong_reader = None


class HL7StreamReader(MLLPStreamReader):
    def __init__(
        self, limit=_DEFAULT_LIMIT, loop=None, encoding=None, encoding_errors=None
    ):
        super().__init__(limit=limit, loop=loop)
        self.encoding = encoding
        self.encoding_errors = encoding_errors

    @property
    def encoding(self):
        return self._encoding

    @encoding.setter
    def encoding(self, encoding):
        assert not encoding or isinstance(encoding, str)
        self._encoding = encoding or "ascii"

    @property
    def encoding_errors(self):
        return self._encoding_errors

    @encoding_errors.setter
    def encoding_errors(self, encoding_errors):
        assert not encoding_errors or isinstance(encoding_errors, str)
        self._encoding_errors = encoding_errors or "strict"

    async def readmessage(self):
        """Reads a full HL7 message from the stream.

        This will return an :py:class:`hl7.Message`.

        If `limit` is reached, `ValueError` will be raised. In that case, if
        block termination separator was found, complete line including separator
        will be removed from internal buffer. Else, internal buffer will be cleared. Limit is
        compared against part of the line without separator.

        If an invalid MLLP block is encountered, :py:class:`hl7.mllp.InvalidBlockError` will be
        raised.
        """
        block = await self.readblock()
        return hl7_parse(block.decode(self.encoding, self.encoding_errors))


class HL7StreamWriter(MLLPStreamWriter):
    def __init__(
        self, transport, protocol, reader, loop, encoding=None, encoding_errors=None
    ):
        super().__init__(transport, protocol, reader, loop)
        self.encoding = encoding
        self.encoding_errors = encoding_errors

    @property
    def encoding(self):
        return self._encoding

    @encoding.setter
    def encoding(self, encoding):
        assert not encoding or isinstance(encoding, str)
        self._encoding = encoding or "ascii"

    @property
    def encoding_errors(self):
        return self._encoding_errors

    @encoding_errors.setter
    def encoding_errors(self, encoding_errors):
        assert not encoding_errors or isinstance(encoding_errors, str)
        self._encoding_errors = encoding_errors or "strict"

    def writemessage(self, message):
        """Writes an :py:class:`hl7.Message` to the stream.
        """
        self.writeblock(str(message).encode(self.encoding, self.encoding_errors))



import asyncio
import asyncio.streams
import sys
from unittest.mock import create_autospec

import hl7
import hl7.mllp

# IsolatedAsyncioTestCase added in 3.8, use backport for older code
if sys.version_info.major <= 3 and sys.version_info.minor < 8:
    from .backports.unittest.async_case import IsolatedAsyncioTestCase
else:
    from unittest import IsolatedAsyncioTestCase


START_BLOCK = b"\x0b"
END_BLOCK = b"\x1c"
CARRIAGE_RETURN = b"\x0d"


class MLLPStreamWriterTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.transport = create_autospec(asyncio.Transport)

    async def asyncSetUp(self):
        self.writer = hl7.mllp.MLLPStreamWriter(
            self.transport,
            create_autospec(asyncio.streams.StreamReaderProtocol),
            create_autospec(hl7.mllp.MLLPStreamReader),
            asyncio.get_running_loop(),
        )

    def test_writeblock(self):
        self.writer.writeblock(b"foobar")
        self.transport.write.assert_called_with(
            START_BLOCK + b"foobar" + END_BLOCK + CARRIAGE_RETURN
        )


class MLLPStreamReaderTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.reader = hl7.mllp.MLLPStreamReader()

    async def test_readblock(self):
        self.reader.feed_data(START_BLOCK + b"foobar" + END_BLOCK + CARRIAGE_RETURN)
        block = await self.reader.readblock()
        self.assertEqual(block, b"foobar")


class HL7StreamWriterTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.transport = create_autospec(asyncio.Transport)

    async def asyncSetUp(self):
        def mock_cb(reader, writer):
            pass

        reader = create_autospec(asyncio.streams.StreamReader)

        self.writer = hl7.mllp.HL7StreamWriter(
            self.transport,
            hl7.mllp.HL7StreamProtocol(reader, mock_cb, asyncio.get_running_loop()),
            reader,
            asyncio.get_running_loop(),
        )

    def test_writemessage(self):
        message = r"MSH|^~\&|LABADT|DH|EPICADT|DH|201301011228||ACK^A01^ACK|HL7ACK00001|P|2.3\r"
        message += "MSA|AA|HL7MSG00001\r"
        hl7_message = hl7.parse(message)
        self.writer.writemessage(hl7_message)
        self.transport.write.assert_called_with(
            START_BLOCK + message.encode() + END_BLOCK + CARRIAGE_RETURN
        )


class HL7StreamReaderTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.reader = hl7.mllp.HL7StreamReader()

    async def test_readblock(self):
        message = r"MSH|^~\&|LABADT|DH|EPICADT|DH|201301011228||ACK^A01^ACK|HL7ACK00001|P|2.3\r"
        message += "MSA|AA|HL7MSG00001\r"
        self.reader.feed_data(
            START_BLOCK + message.encode() + END_BLOCK + CARRIAGE_RETURN
        )
        hl7_message = await self.reader.readmessage()
        self.assertEqual(str(hl7_message), str(hl7.parse(message)))



# -*- coding: utf-8 -*-
from unittest import TestCase

from hl7 import Accessor


class AccessorTest(TestCase):
    def test_key(self):
        self.assertEqual("FOO", Accessor("FOO").key)
        self.assertEqual("FOO2", Accessor("FOO", 2).key)
        self.assertEqual("FOO2.3", Accessor("FOO", 2, 3).key)
        self.assertEqual("FOO2.3.1.4.6", Accessor("FOO", 2, 3, 1, 4, 6).key)

    def test_parse(self):
        self.assertEqual(Accessor("FOO"), Accessor.parse_key("FOO"))
        self.assertEqual(
            Accessor("FOO", 2, 3, 1, 4, 6), Accessor.parse_key("FOO2.3.1.4.6")
        )

    def test_equality(self):
        self.assertEqual(Accessor("FOO", 1, 3, 4), Accessor("FOO", 1, 3, 4))
        self.assertNotEqual(Accessor("FOO", 1), Accessor("FOO", 2))



# -*- coding: utf-8 -*-
import datetime
import logging
import random
import string

logger = logging.getLogger(__file__)


def ishl7(line):
    """Determines whether a *line* looks like an HL7 message.
    This method only does a cursory check and does not fully
    validate the message.

    :rtype: bool
    """
    # Prevent issues if the line is empty
    return line and (line.strip()[:3] in ["MSH"]) or False


def isfile(line):
    """
        Files are wrapped in FHS / FTS
        FHS = file header segment
        FTS = file trailer segment
    """
    return line and (line.strip()[:3] in ["FHS"]) or False


def split_file(hl7file):
    """
        Given a file, split out the messages.
        Does not do any validation on the message.
        Throws away batch and file segments.
    """
    rv = []
    for line in hl7file.split("\r"):
        line = line.strip()
        if line[:3] in ["FHS", "BHS", "FTS", "BTS"]:
            continue
        if line[:3] == "MSH":
            newmsg = [line]
            rv.append(newmsg)
        else:
            if len(rv) == 0:
                logger.error("Segment received before message header [%s]", line)
                continue
            rv[-1].append(line)
    rv = ["\r".join(msg) for msg in rv]
    for i, msg in enumerate(rv):
        if not msg[-1] == "\r":
            rv[i] = msg + "\r"
    return rv


alphanumerics = string.ascii_uppercase + string.digits


def generate_message_control_id():
    """Generate a unique 20 character message id.

    See http://www.hl7resources.com/Public/index.html?a55433.htm
    """
    d = datetime.datetime.utcnow()
    # Strip off the decade, ID only has to be unique for 3 years.
    # So now we have a 16 char timestamp.
    timestamp = d.strftime("%y%j%H%M%S%f")[1:]
    # Add 4 chars of uniqueness
    unique = "".join(random.sample(alphanumerics, 4))
    return timestamp + unique



from datetime import datetime
from unittest import TestCase

from hl7.datatypes import _UTCOffset, parse_datetime


class DatetimeTest(TestCase):
    def test_parse_date(self):
        self.assertEqual(datetime(1901, 2, 13), parse_datetime("19010213"))

    def test_parse_datetime(self):
        self.assertEqual(
            datetime(2014, 3, 11, 14, 25, 33), parse_datetime("20140311142533")
        )

    def test_parse_datetime_frac(self):
        self.assertEqual(
            datetime(2014, 3, 11, 14, 25, 33, 100000),
            parse_datetime("20140311142533.1"),
        )
        self.assertEqual(
            datetime(2014, 3, 11, 14, 25, 33, 10000),
            parse_datetime("20140311142533.01"),
        )
        self.assertEqual(
            datetime(2014, 3, 11, 14, 25, 33, 1000),
            parse_datetime("20140311142533.001"),
        )
        self.assertEqual(
            datetime(2014, 3, 11, 14, 25, 33, 100),
            parse_datetime("20140311142533.0001"),
        )

    def test_parse_tz(self):
        self.assertEqual(
            datetime(2014, 3, 11, 14, 12, tzinfo=_UTCOffset(330)),
            parse_datetime("201403111412+0530"),
        )
        self.assertEqual(
            datetime(2014, 3, 11, 14, 12, 20, tzinfo=_UTCOffset(-300)),
            parse_datetime("20140311141220-0500"),
        )

    def test_tz(self):
        self.assertEqual("+0205", _UTCOffset(125).tzname(datetime.utcnow()))
        self.assertEqual("-0410", _UTCOffset(-250).tzname(datetime.utcnow()))



# -*- coding: utf-8 -*-

"""
Primary version number source.

Forth element can be 'dev' < 'a' < 'b' < 'rc' < 'final'. An empty 4th
element is equivalent to 'final'.
"""
VERSION = (0, 4, 1, "final")


def get_version():
    """Provide version number

    Use verlib format [1]_:
      N.N[.N]+[{a|b|c|rc}N[.N]+][.postN][.devN]

    .. [1] http://www.python.org/dev/peps/pep-0386/
    """
    main_version = "%s.%s.%s" % VERSION[0:3]

    if len(VERSION) < 4:
        return main_version

    version_type = VERSION[3]
    if not version_type or version_type == "final":
        return main_version
    elif version_type == "dev":
        return "%s.dev" % main_version
    else:
        return "%s%s" % (main_version, version_type)



# -*- coding: utf-8 -*-
from .containers import Factory


def parse(line, encoding="utf-8", factory=Factory):
    """Returns a instance of the :py:class:`hl7.Message` that allows
    indexed access to the data elements.

    A custom :py:class:`hl7.Factory` subclass can be passed in to be used when
    constructing the message and it's components.

    .. note::

        HL7 usually contains only ASCII, but can use other character
        sets (HL7 Standards Document, Section 1.7.1), however as of v2.8,
        UTF-8 is the preferred character set [#]_.

        python-hl7 works on Python unicode strings. :py:func:`hl7.parse`
        will accept unicode string or will attempt to convert bytestrings
        into unicode strings using the optional ``encoding`` parameter.
        ``encoding`` defaults to UTF-8, so no work is needed for bytestrings
        in UTF-8, but for other character sets like 'cp1252' or 'latin1',
        ``encoding`` must be set appropriately.

    >>> h = hl7.parse(message)

    To decode a non-UTF-8 byte string::

       hl7.parse(message, encoding='latin1')

    :rtype: :py:class:`hl7.Message`

    .. [#] http://wiki.hl7.org/index.php?title=Character_Set_used_in_v2_messages

    """
    # Ensure we are working with unicode data, decode the bytestring
    # if needed
    if isinstance(line, bytes):
        line = line.decode(encoding)
    # Strip out unnecessary whitespace
    strmsg = line.strip()
    # The method for parsing the message
    plan = create_parse_plan(strmsg, factory)
    # Start spliting the methods based upon the ParsePlan
    return _split(strmsg, plan)


def _split(text, plan):
    """Recursive function to split the *text* into an n-deep list,
    according to the :py:class:`hl7._ParsePlan`.
    """
    # Base condition, if we have used up all the plans
    if not plan:
        return text

    if not plan.applies(text):
        return plan.container([text])

    # Parsing of the first segment is awkward because it contains
    # the separator characters in a field
    if plan.containers[0] == plan.factory.create_segment and text[:3] in ["MSH", "FHS"]:
        seg = text[:3]
        sep0 = text[3]
        sep_end_off = text.find(sep0, 4)
        seps = text[4:sep_end_off]
        text = text[sep_end_off + 1 :]
        data = [
            plan.factory.create_field("", [seg]),
            plan.factory.create_field("", [sep0]),
            plan.factory.create_field(sep0, [seps]),
        ]
    else:
        data = []

    if text:
        data = data + [_split(x, plan.next()) for x in text.split(plan.separator)]
    # Return the instance of the current message part according
    # to the plan
    return plan.container(data)


def create_parse_plan(strmsg, factory=Factory):
    """Creates a plan on how to parse the HL7 message according to
    the details stored within the message.
    """
    # We will always use a carriage return to separate segments
    separators = ["\r"]

    # Extract the rest of the separators. Defaults used if not present.
    assert strmsg[:3] in ("MSH")
    sep0 = strmsg[3]
    seps = list(strmsg[3 : strmsg.find(sep0, 4)])

    separators.append(seps[0])
    if len(seps) > 2:
        separators.append(seps[2])  # repetition separator
    else:
        separators.append("~")  # repetition separator
    if len(seps) > 1:
        separators.append(seps[1])  # component separator
    else:
        separators.append("^")  # component separator
    if len(seps) > 4:
        separators.append(seps[4])  # sub-component separator
    else:
        separators.append("&")  # sub-component separator
    if len(seps) > 3:
        esc = seps[3]
    else:
        esc = "\\"

    # The ordered list of containers to create
    containers = [
        factory.create_message,
        factory.create_segment,
        factory.create_field,
        factory.create_repetition,
        factory.create_component,
    ]
    return _ParsePlan(separators, containers, esc, factory)


class _ParsePlan(object):
    """Details on how to parse an HL7 message. Typically this object
    should be created via :func:`hl7.create_parse_plan`
    """

    # field, component, repetition, escape, subcomponent

    def __init__(self, separators, containers, esc, factory):
        # TODO test to see performance implications of the assertion
        # since we generate the ParsePlan, this should never be in
        # invalid state
        assert len(containers) == len(separators)
        self.separators = separators
        self.containers = containers
        self.esc = esc
        self.factory = factory

    @property
    def separator(self):
        """Return the current separator to use based on the plan."""
        return self.separators[0]

    def container(self, data):
        """Return an instance of the approriate container for the *data*
        as specified by the current plan.
        """
        return self.containers[0](
            self.separator, data, self.esc, self.separators, self.factory
        )

    def next(self):
        """Generate the next level of the plan (essentially generates
        a copy of this plan with the level of the container and the
        seperator starting at the next index.
        """
        if len(self.containers) > 1:
            # Return a new instance of this class using the tails of
            # the separators and containers lists. Use self.__class__()
            # in case :class:`hl7.ParsePlan` is subclassed
            return self.__class__(
                self.separators[1:], self.containers[1:], self.esc, self.factory
            )
        # When we have no separators and containers left, return None,
        # which indicates that we have nothing further.
        return None

    def applies(self, text):
        """return True if the separator or those if the children are in the text"""
        for s in self.separators:
            if text.find(s) >= 0:
                return True
        return False



# -*- coding: utf-8 -*-
import datetime
import math
import re

DTM_TZ_RE = re.compile(r"(\d+(?:\.\d+)?)(?:([+-]\d{2})(\d{2}))?")


class _UTCOffset(datetime.tzinfo):
    """Fixed offset timezone from UTC."""

    def __init__(self, minutes):
        """``minutes`` is a offset from UTC, negative for west of UTC"""
        self.minutes = minutes

    def utcoffset(self, dt):
        return datetime.timedelta(minutes=self.minutes)

    def tzname(self, dt):
        minutes = abs(self.minutes)
        return "{0}{1:02}{2:02}".format(
            "-" if self.minutes < 0 else "+", minutes // 60, minutes % 60
        )

    def dst(self, dt):
        return datetime.timedelta(0)


def parse_datetime(value):
    """Parse hl7 DTM string ``value`` :py:class:`datetime.datetime`.

    ``value`` is of the format YYYY[MM[DD[HH[MM[SS[.S[S[S[S]]]]]]]]][+/-HHMM]
    or a ValueError will be raised.

    :rtype: :py:;class:`datetime.datetime`
    """
    if not value:
        return None

    # Split off optional timezone
    dt_match = DTM_TZ_RE.match(value)
    if not dt_match:
        raise ValueError("Malformed HL7 datetime {0}".format(value))
    dtm = dt_match.group(1)
    tzh = dt_match.group(2)
    tzm = dt_match.group(3)
    if tzh and tzm:
        minutes = int(tzh) * 60
        minutes += math.copysign(int(tzm), minutes)
        tzinfo = _UTCOffset(minutes)
    else:
        tzinfo = None

    precision = len(dtm)

    if precision >= 4:
        year = int(dtm[0:4])
    else:
        raise ValueError("Malformed HL7 datetime {0}".format(value))

    if precision >= 6:
        month = int(dtm[4:6])
    else:
        month = 1

    if precision >= 8:
        day = int(dtm[6:8])
    else:
        day = 1

    if precision >= 10:
        hour = int(dtm[8:10])
    else:
        hour = 0

    if precision >= 12:
        minute = int(dtm[10:12])
    else:
        minute = 0

    if precision >= 14:
        delta = datetime.timedelta(seconds=float(dtm[12:]))
        second = delta.seconds
        microsecond = delta.microseconds
    else:
        second = 0
        microsecond = 0

    return datetime.datetime(
        year, month, day, hour, minute, second, microsecond, tzinfo=tzinfo
    )



# -*- coding: utf-8 -*-
from unittest import TestCase

import hl7

from .samples import rep_sample_hl7

SEP = r"|^~\&"
CR_SEP = "\r"


class ConstructionTest(TestCase):
    def test_create_msg(self):
        # Create a message
        MSH = hl7.Segment(SEP[0], [hl7.Field(SEP[1], ["MSH"])])
        MSA = hl7.Segment(SEP[0], [hl7.Field(SEP[1], ["MSA"])])
        response = hl7.Message(CR_SEP, [MSH, MSA])
        response["MSH.F1.R1"] = SEP[0]
        response["MSH.F2.R1"] = SEP[1:]
        self.assertEqual(str(response), "MSH|^~\\&|\rMSA\r")

    def test_append(self):
        # Append a segment to a message
        MSH = hl7.Segment(SEP[0], [hl7.Field(SEP[1], ["MSH"])])
        response = hl7.Message(CR_SEP, [MSH])
        response["MSH.F1.R1"] = SEP[0]
        response["MSH.F2.R1"] = SEP[1:]
        MSA = hl7.Segment(SEP[0], [hl7.Field(SEP[1], ["MSA"])])
        response.append(MSA)
        self.assertEqual(str(response), "MSH|^~\\&|\rMSA\r")

    def test_append_from_source(self):
        # Copy a segment between messages
        MSH = hl7.Segment(SEP[0], [hl7.Field(SEP[1], ["MSH"])])
        MSA = hl7.Segment(SEP[0], [hl7.Field(SEP[1], ["MSA"])])
        response = hl7.Message(CR_SEP, [MSH, MSA])
        response["MSH.F1.R1"] = SEP[0]
        response["MSH.F2.R1"] = SEP[1:]
        self.assertEqual(str(response), "MSH|^~\\&|\rMSA\r")
        src_msg = hl7.parse(rep_sample_hl7)
        PID = src_msg["PID"][0]
        self.assertEqual(
            str(PID),
            "PID|Field1|Component1^Component2|Component1^Sub-Component1&Sub-Component2^Component3|Repeat1~Repeat2",
        )
        response.append(PID)
        self.assertEqual(
            str(response),
            "MSH|^~\\&|\rMSA\rPID|Field1|Component1^Component2|Component1^Sub-Component1&Sub-Component2^Component3|Repeat1~Repeat2\r",
        )



# -*- coding: utf-8 -*-
"""python-hl7 is a simple library for parsing messages of Health Level 7
(HL7) version 2.x into Python objects.

* Documentation: http://python-hl7.readthedocs.org
* Source Code: http://github.com/johnpaulett/python-hl7
"""
from .accessor import Accessor
from .containers import (
    Component,
    Container,
    Factory,
    Field,
    Message,
    Repetition,
    Segment,
    Sequence,
)
from .datatypes import parse_datetime
from .parser import parse
from .util import generate_message_control_id, isfile, ishl7, split_file
from .version import get_version

__version__ = get_version()
__author__ = "John Paulett"
__email__ = "john -at- paulett.org"
__license__ = "BSD"
__copyright__ = "Copyright 2011, John Paulett <john -at- paulett.org>"

#: This is the HL7 Null value. It means that a field is present and blank.
NULL = '""'


__all__ = [
    "parse",
    "Sequence",
    "Container",
    "Message",
    "Segment",
    "Field",
    "Repetition",
    "Component",
    "Factory",
    "Accessor",
    "ishl7",
    "isfile",
    "split_file",
    "generate_message_control_id",
    "parse_datetime",
]



import os
import socket
from optparse import Values
from shutil import rmtree
from tempfile import mkdtemp
from unittest import TestCase
from unittest.mock import Mock, patch

import hl7
from hl7 import __version__ as hl7_version
from hl7.client import CR, EB, SB, MLLPClient, MLLPException, mllp_send


class MLLPClientTest(TestCase):
    def setUp(self):
        # use a mock version of socket
        self.socket_patch = patch("hl7.client.socket.socket")
        self.mock_socket = self.socket_patch.start()

        self.client = MLLPClient("localhost", 6666)

    def tearDown(self):
        # unpatch socket
        self.socket_patch.stop()

    def test_connect(self):
        self.mock_socket.assert_called_once_with(socket.AF_INET, socket.SOCK_STREAM)
        self.client.socket.connect.assert_called_once_with(("localhost", 6666))

    def test_close(self):
        self.client.close()
        self.client.socket.close.assert_called_once_with()

    def test_send(self):
        self.client.socket.recv.return_value = "thanks"

        result = self.client.send("foobar\n")
        self.assertEqual(result, "thanks")

        self.client.socket.send.assert_called_once_with("foobar\n")
        self.client.socket.recv.assert_called_once_with(4096)

    def test_send_message_unicode(self):
        self.client.socket.recv.return_value = "thanks"

        result = self.client.send_message(u"foobar")
        self.assertEqual(result, "thanks")

        self.client.socket.send.assert_called_once_with(b"\x0bfoobar\x1c\x0d")

    def test_send_message_bytestring(self):
        self.client.socket.recv.return_value = "thanks"

        result = self.client.send_message(b"foobar")
        self.assertEqual(result, "thanks")

        self.client.socket.send.assert_called_once_with(b"\x0bfoobar\x1c\x0d")

    def test_send_message_hl7_message(self):
        self.client.socket.recv.return_value = "thanks"

        message = hl7.parse(r"MSH|^~\&|GHH LAB|ELAB")

        result = self.client.send_message(message)
        self.assertEqual(result, "thanks")

        self.client.socket.send.assert_called_once_with(
            b"\x0bMSH|^~\\&|GHH LAB|ELAB\r\x1c\x0d"
        )

    def test_context_manager(self):
        with MLLPClient("localhost", 6666) as client:
            client.send("hello world")

        self.client.socket.send.assert_called_once_with("hello world")
        self.client.socket.close.assert_called_once_with()

    def test_context_manager_exception(self):
        with self.assertRaises(Exception):
            with MLLPClient("localhost", 6666):
                raise Exception()

        # socket.close should be called via the with statement
        self.client.socket.close.assert_called_once_with()


class MLLPSendTest(TestCase):
    def setUp(self):
        # patch to avoid touching sys and socket
        self.socket_patch = patch("hl7.client.socket.socket")
        self.mock_socket = self.socket_patch.start()
        self.mock_socket().recv.return_value = "thanks"

        self.stdout_patch = patch("hl7.client.stdout")
        self.mock_stdout = self.stdout_patch.start()

        self.stdin_patch = patch("hl7.client.stdin")
        self.mock_stdin = self.stdin_patch.start()

        self.stderr_patch = patch("hl7.client.stderr")
        self.mock_stderr = self.stderr_patch.start()

        self.exit_patch = patch("hl7.client.sys.exit")
        self.mock_exit = self.exit_patch.start()

        # we need a temporary directory
        self.dir = mkdtemp()
        self.write(SB + b"foobar" + EB + CR)

        self.option_values = Values(
            {
                "port": 6661,
                "filename": os.path.join(self.dir, "test.hl7"),
                "verbose": True,
                "loose": False,
                "version": False,
            }
        )

        self.options_patch = patch("hl7.client.OptionParser")
        option_parser = self.options_patch.start()
        self.mock_options = Mock()
        option_parser.return_value = self.mock_options
        self.mock_options.parse_args.return_value = (self.option_values, ["localhost"])

    def tearDown(self):
        # unpatch
        self.socket_patch.stop()
        self.options_patch.stop()
        self.stdout_patch.stop()
        self.stdin_patch.stop()
        self.stderr_patch.stop()
        self.exit_patch.stop()

        # clean up the temp directory
        rmtree(self.dir)

    def write(self, content, path="test.hl7"):
        with open(os.path.join(self.dir, path), "wb") as f:
            f.write(content)

    def test_send(self):
        mllp_send()

        self.mock_socket().connect.assert_called_once_with(("localhost", 6661))
        self.mock_socket().send.assert_called_once_with(SB + b"foobar" + EB + CR)
        self.mock_stdout.assert_called_once_with("thanks")
        self.assertFalse(self.mock_exit.called)

    def test_send_multiple(self):
        self.mock_socket().recv.return_value = "thanks"
        self.write(SB + b"foobar" + EB + CR + SB + b"hello" + EB + CR)

        mllp_send()

        self.assertEqual(
            self.mock_socket().send.call_args_list[0][0][0], SB + b"foobar" + EB + CR
        )
        self.assertEqual(
            self.mock_socket().send.call_args_list[1][0][0], SB + b"hello" + EB + CR
        )

    def test_leftover_buffer(self):
        self.write(SB + b"foobar" + EB + CR + SB + b"stuff")

        self.assertRaises(MLLPException, mllp_send)

        self.mock_socket().send.assert_called_once_with(SB + b"foobar" + EB + CR)

    def test_quiet(self):
        self.option_values.verbose = False

        mllp_send()

        self.mock_socket().send.assert_called_once_with(SB + b"foobar" + EB + CR)
        self.assertFalse(self.mock_stdout.called)

    def test_port(self):
        self.option_values.port = 7890

        mllp_send()

        self.mock_socket().connect.assert_called_once_with(("localhost", 7890))

    def test_stdin(self):
        self.option_values.filename = None
        self.mock_stdin.return_value = FakeStream()

        mllp_send()

        self.mock_socket().send.assert_called_once_with(SB + b"hello" + EB + CR)

    def test_loose_no_stdin(self):
        self.option_values.loose = True
        self.option_values.filename = None
        self.mock_stdin.return_value = FakeStream()

        mllp_send()

        self.assertFalse(self.mock_socket().send.called)
        self.mock_stderr().write.assert_called_with("--loose requires --file\n")
        self.mock_exit.assert_called_with(1)

    def test_loose_windows_newline(self):
        self.option_values.loose = True
        self.write(SB + b"MSH|^~\\&|foo\r\nbar\r\n" + EB + CR)

        mllp_send()

        self.mock_socket().send.assert_called_once_with(
            SB + b"MSH|^~\\&|foo\rbar" + EB + CR
        )

    def test_loose_unix_newline(self):
        self.option_values.loose = True
        self.write(SB + b"MSH|^~\\&|foo\nbar\n" + EB + CR)

        mllp_send()

        self.mock_socket().send.assert_called_once_with(
            SB + b"MSH|^~\\&|foo\rbar" + EB + CR
        )

    def test_loose_no_mllp_characters(self):
        self.option_values.loose = True
        self.write(b"MSH|^~\\&|foo\r\nbar\r\n")

        mllp_send()

        self.mock_socket().send.assert_called_once_with(
            SB + b"MSH|^~\\&|foo\rbar" + EB + CR
        )

    def test_loose_send_mutliple(self):
        self.option_values.loose = True
        self.mock_socket().recv.return_value = "thanks"
        self.write(b"MSH|^~\\&|1\r\nOBX|1\r\nMSH|^~\\&|2\r\nOBX|2\r\n")

        mllp_send()

        self.assertEqual(
            self.mock_socket().send.call_args_list[0][0][0],
            SB + b"MSH|^~\\&|1\rOBX|1" + EB + CR,
        )
        self.assertEqual(
            self.mock_socket().send.call_args_list[1][0][0],
            SB + b"MSH|^~\\&|2\rOBX|2" + EB + CR,
        )

    def test_version(self):
        self.option_values.version = True

        mllp_send()

        self.assertFalse(self.mock_socket().connect.called)
        self.mock_stdout.assert_called_once_with(str(hl7_version))


class FakeStream(object):
    count = 0

    def read(self, buf):
        self.count += 1
        if self.count == 1:
            return SB + b"hello" + EB + CR
        else:
            return b""



# -*- coding: utf-8 -*-
from unittest import TestCase

import hl7
from hl7 import Accessor, Component, Field, Message, Repetition, Segment

from .samples import rep_sample_hl7, sample_file, sample_hl7


class ParseTest(TestCase):
    def test_parse(self):
        msg = hl7.parse(sample_hl7)
        self.assertEqual(len(msg), 5)
        self.assertIsInstance(msg[0][0][0], str)
        self.assertEqual(msg[0][0][0], "MSH")
        self.assertEqual(msg[3][0][0], "OBX")
        self.assertEqual(
            msg[3][3],
            [[["1554-5"], ["GLUCOSE"], ["POST 12H CFST:MCNC:PT:SER/PLAS:QN"]]],
        )
        # Make sure MSH-1 and MSH-2 are valid
        self.assertEqual(msg[0][1][0], "|")
        self.assertIsInstance(msg[0][1], hl7.Field)
        self.assertEqual(msg[0][2][0], r"^~\&")
        self.assertIsInstance(msg[0][2], hl7.Field)
        # MSH-9 is the message type
        self.assertEqual(msg[0][9], [[["ORU"], ["R01"]]])
        # Do it twice to make sure text coercion is idempotent
        self.assertEqual(str(msg), sample_hl7)
        self.assertEqual(str(msg), sample_hl7)

    def test_bytestring_converted_to_unicode(self):
        msg = hl7.parse(str(sample_hl7))
        self.assertEqual(len(msg), 5)
        self.assertIsInstance(msg[0][0][0], str)
        self.assertEqual(msg[0][0][0], "MSH")

    def test_non_ascii_bytestring(self):
        # \x96 - valid cp1252, not valid utf8
        # it is the responsibility of the caller to convert to unicode
        msg = hl7.parse(b"MSH|^~\\&|GHH LAB|ELAB\x963", encoding="cp1252")
        self.assertEqual(msg[0][4][0], "ELAB\u20133")

    def test_non_ascii_bytestring_no_encoding(self):
        # \x96 - valid cp1252, not valid utf8
        # it is the responsibility of the caller to convert to unicode
        self.assertRaises(UnicodeDecodeError, hl7.parse, b"MSH|^~\\&|GHH LAB|ELAB\x963")

    def test_parsing_classes(self):
        msg = hl7.parse(sample_hl7)

        self.assertIsInstance(msg, hl7.Message)
        self.assertIsInstance(msg[3], hl7.Segment)
        self.assertIsInstance(msg[3][0], hl7.Field)
        self.assertIsInstance(msg[3][0][0], str)

    def test_nonstandard_separators(self):
        nonstd = "MSH$%~\\&$GHH LAB\rPID$$$555-44-4444$$EVERYWOMAN%EVE%E%%%L\r"
        msg = hl7.parse(nonstd)

        self.assertEqual(str(msg), nonstd)
        self.assertEqual(len(msg), 2)
        self.assertEqual(
            msg[1][5], [[["EVERYWOMAN"], ["EVE"], ["E"], [""], [""], ["L"]]]
        )

    def test_repetition(self):
        msg = hl7.parse(rep_sample_hl7)
        self.assertEqual(msg[1][4], [["Repeat1"], ["Repeat2"]])
        self.assertIsInstance(msg[1][4], Field)
        self.assertIsInstance(msg[1][4][0], Repetition)
        self.assertIsInstance(msg[1][4][1], Repetition)
        self.assertEqual(str(msg[1][4][0][0]), "Repeat1")
        self.assertIsInstance(msg[1][4][0][0], str)
        self.assertEqual(str(msg[1][4][1][0]), "Repeat2")
        self.assertIsInstance(msg[1][4][1][0], str)

    def test_empty_initial_repetition(self):
        # Switch to look like "|~Repeat2|
        msg = hl7.parse(rep_sample_hl7.replace("Repeat1", ""))
        self.assertEqual(msg[1][4], [[""], ["Repeat2"]])

    def test_subcomponent(self):
        msg = hl7.parse(rep_sample_hl7)
        self.assertEqual(
            msg[1][3],
            [[["Component1"], ["Sub-Component1", "Sub-Component2"], ["Component3"]]],
        )

    def test_elementnumbering(self):
        # Make sure that the numbering of repetitions. components and
        # sub-components is indexed from 1 when invoked as callable
        # (for compatibility with HL7 spec numbering)
        # and not 0-based (default for Python list)
        msg = hl7.parse(rep_sample_hl7)
        f = msg(2)(3)(1)(2)(2)
        self.assertIs(f, msg["PID.3.1.2.2"])
        self.assertIs(f, msg[1][3][0][1][1])
        f = msg(2)(4)(2)(1)
        self.assertIs(f, msg["PID.4.2.1"])
        self.assertIs(f, msg[1][4][1][0])
        # Repetition level accessed in list-form doesn't make much sense...
        self.assertIs(f, msg["PID.4.2"])

    def test_extract(self):
        msg = hl7.parse(rep_sample_hl7)

        # Full correct path
        self.assertEqual(msg["PID.3.1.2.2"], "Sub-Component2")
        self.assertEqual(msg[Accessor("PID", 1, 3, 1, 2, 2)], "Sub-Component2")

        # Shorter Paths
        self.assertEqual(msg["PID.1.1"], "Field1")
        self.assertEqual(msg[Accessor("PID", 1, 1, 1)], "Field1")
        self.assertEqual(msg["PID.1"], "Field1")
        self.assertEqual(msg["PID1.1"], "Field1")
        self.assertEqual(msg["PID.3.1.2"], "Sub-Component1")

        # Longer Paths
        self.assertEqual(msg["PID.1.1.1.1"], "Field1")

        # Incorrect path
        self.assertRaisesRegex(
            IndexError,
            "PID.1.1.1.2",
            msg.extract_field,
            *Accessor.parse_key("PID.1.1.1.2")
        )

        # Optional field, not included in message
        self.assertEqual(msg["MSH.20"], "")

        # Optional sub-component, not included in message
        self.assertEqual(msg["PID.3.1.2.3"], "")
        self.assertEqual(msg["PID.3.1.3"], "Component3")
        self.assertEqual(msg["PID.3.1.4"], "")

    def test_assign(self):
        msg = hl7.parse(rep_sample_hl7)

        # Field
        msg["MSH.20"] = "FIELD 20"
        self.assertEqual(msg["MSH.20"], "FIELD 20")

        # Component
        msg["MSH.21.1.1"] = "COMPONENT 21.1.1"
        self.assertEqual(msg["MSH.21.1.1"], "COMPONENT 21.1.1")

        # Sub-Component
        msg["MSH.21.1.2.4"] = "SUBCOMPONENT 21.1.2.4"
        self.assertEqual(msg["MSH.21.1.2.4"], "SUBCOMPONENT 21.1.2.4")

        # Verify round-tripping (i.e. that separators are correct)
        msg2 = hl7.parse(str(msg))
        self.assertEqual(msg2["MSH.20"], "FIELD 20")
        self.assertEqual(msg2["MSH.21.1.1"], "COMPONENT 21.1.1")
        self.assertEqual(msg2["MSH.21.1.2.4"], "SUBCOMPONENT 21.1.2.4")

    def test_unescape(self):
        msg = hl7.parse(rep_sample_hl7)

        # Replace Separators
        self.assertEqual(msg.unescape("\\E\\"), "\\")
        self.assertEqual(msg.unescape("\\F\\"), "|")
        self.assertEqual(msg.unescape("\\S\\"), "^")
        self.assertEqual(msg.unescape("\\T\\"), "&")
        self.assertEqual(msg.unescape("\\R\\"), "~")

        # Replace Highlighting
        self.assertEqual(msg.unescape("\\H\\text\\N\\"), "_text_")

        # Application Overrides
        self.assertEqual(msg.unescape("\\H\\text\\N\\", {"H": "*", "N": "*"}), "*text*")

        # Hex Codes
        self.assertEqual(msg.unescape("\\X20202020\\"), "    ")
        self.assertEqual(msg.unescape("\\Xe1\\\\Xe9\\\\Xed\\\\Xf3\\\\Xfa\\"), "áéíóú")

    def test_escape(self):
        msg = hl7.parse(rep_sample_hl7)

        # Escape Separators
        self.assertEqual(msg.escape("\\"), "\\E\\")
        self.assertEqual(msg.escape("|"), "\\F\\")
        self.assertEqual(msg.escape("^"), "\\S\\")
        self.assertEqual(msg.escape("&"), "\\T\\")
        self.assertEqual(msg.escape("~"), "\\R\\")

        # Escape ASCII characters
        self.assertEqual(msg.escape("asdf"), "asdf")

        # Escape non-ASCII characters
        self.assertEqual(msg.escape("áéíóú"), "\\Xe1\\\\Xe9\\\\Xed\\\\Xf3\\\\Xfa\\")
        self.assertEqual(msg.escape("äsdf"), "\\Xe4\\sdf")

    def test_file(self):
        # Extract message from file
        self.assertTrue(hl7.isfile(sample_file))
        messages = hl7.split_file(sample_file)
        self.assertEqual(len(messages), 1)

        # message can be parsed
        msg = hl7.parse(messages[0])

        # message has expected content
        self.assertEqual(
            [s[0][0] for s in msg], ["MSH", "EVN", "PID", "PD1", "NK1", "PV1"]
        )


class ParsePlanTest(TestCase):
    def test_create_parse_plan(self):
        plan = hl7.parser.create_parse_plan(sample_hl7)

        self.assertEqual(plan.separators, ["\r", "|", "~", "^", "&"])
        self.assertEqual(
            plan.containers, [Message, Segment, Field, Repetition, Component]
        )

    def test_parse_plan(self):
        plan = hl7.parser.create_parse_plan(sample_hl7)

        self.assertEqual(plan.separator, "\r")
        con = plan.container([1, 2])
        self.assertIsInstance(con, Message)
        self.assertEqual(con, [1, 2])
        self.assertEqual(con.separator, "\r")

    def test_parse_plan_next(self):
        plan = hl7.parser.create_parse_plan(sample_hl7)

        n1 = plan.next()
        self.assertEqual(n1.separators, ["|", "~", "^", "&"])
        self.assertEqual(n1.containers, [Segment, Field, Repetition, Component])

        n2 = n1.next()
        self.assertEqual(n2.separators, ["~", "^", "&"])
        self.assertEqual(n2.containers, [Field, Repetition, Component])

        n3 = n2.next()
        self.assertEqual(n3.separators, ["^", "&"])
        self.assertEqual(n3.containers, [Repetition, Component])

        n4 = n3.next()
        self.assertEqual(n4.separators, ["&"])
        self.assertEqual(n4.containers, [Component])

        n5 = n4.next()
        self.assertTrue(n5 is None)



# -*- coding: utf-8 -*-
from unittest import TestCase

import hl7

from .samples import sample_file, sample_hl7


class IsHL7Test(TestCase):
    def test_ishl7(self):
        self.assertTrue(hl7.ishl7(sample_hl7))

    def test_ishl7_empty(self):
        self.assertFalse(hl7.ishl7(""))

    def test_ishl7_None(self):
        self.assertFalse(hl7.ishl7(None))

    def test_ishl7_wrongsegment(self):
        message = "OBX|1|SN|1554-5^GLUCOSE^POST 12H CFST:MCNC:PT:SER/PLAS:QN||^182|mg/dl|70_105|H|||F\r"
        self.assertFalse(hl7.ishl7(message))

    def test_isfile(self):
        self.assertFalse(hl7.ishl7(sample_file))
        self.assertTrue(hl7.isfile(sample_file))



from .exceptions import InvalidBlockError
from .streams import (
    HL7StreamProtocol,
    HL7StreamReader,
    HL7StreamWriter,
    MLLPStreamReader,
    MLLPStreamWriter,
    open_hl7_connection,
    start_hl7_server,
)

__all__ = [
    "open_hl7_connection",
    "start_hl7_server",
    "HL7StreamProtocol",
    "HL7StreamReader",
    "HL7StreamWriter",
    "MLLPStreamReader",
    "MLLPStreamWriter",
    "InvalidBlockError",
]



# -*- coding: utf-8 -*-

from setuptools import setup

# Avoid directly importing the module. Prevents potential circular
# references when dependency needs to be installed via setup.py, so it
# is not yet available to setup.py
exec(open("hl7/version.py").read())

setup(
    name="hl7",
    version=get_version(),  # noqa
    description="Python library parsing HL7 v2.x messages",
    long_description="""
    python-hl7 is a simple library for parsing messages of Health Level 7
    (HL7) version 2.x into Python objects.

    * Documentation: http://python-hl7.readthedocs.org
    * Source Code: http://github.com/johnpaulett/python-hl7
    """,
    author="John Paulett",
    author_email="john@paulett.org",
    url="http://python-hl7.readthedocs.org",
    license="BSD",
    platforms=["POSIX", "Windows"],
    keywords=[
        "HL7",
        "Health Level 7",
        "healthcare",
        "health care",
        "medical record",
        "mllp",
    ],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Healthcare Industry",
        "Topic :: Communications",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=["hl7", "hl7.mllp"],
    install_requires=[],
    test_suite="tests",
    tests_require=[],
    entry_points={"console_scripts": ["mllp_send=hl7.client:mllp_send",],},
    zip_safe=True,
)



from unittest import TestCase
from unittest.mock import patch

from hl7.version import get_version


class GetVersionTest(TestCase):
    @patch("hl7.version.VERSION", new=(0, 4, 1))
    def test_no_modifier(self):
        self.assertEqual("0.4.1", get_version())

    @patch("hl7.version.VERSION", new=(0, 4, 1, ""))
    def test_empty_modifier(self):
        self.assertEqual("0.4.1", get_version())

    @patch("hl7.version.VERSION", new=(0, 4, 1, None))
    def test_none_modifier(self):
        self.assertEqual("0.4.1", get_version())

    @patch("hl7.version.VERSION", new=(0, 4, 1, "final"))
    def test_final(self):
        self.assertEqual("0.4.1", get_version())

    @patch("hl7.version.VERSION", new=(0, 4, 1, "rc"))
    def test_rc(self):
        self.assertEqual("0.4.1rc", get_version())

    @patch("hl7.version.VERSION", new=(0, 4, 1, "rc4"))
    def test_rc_num(self):
        self.assertEqual("0.4.1rc4", get_version())

    @patch("hl7.version.VERSION", new=(0, 4, 1, "b"))
    def test_beta(self):
        self.assertEqual("0.4.1b", get_version())

    @patch("hl7.version.VERSION", new=(0, 4, 1, "a"))
    def test_alpha(self):
        self.assertEqual("0.4.1a", get_version())

    @patch("hl7.version.VERSION", new=(0, 4, 1, "dev"))
    def test_dev(self):
        self.assertEqual("0.4.1.dev", get_version())



# -*- coding: utf-8 -*-
import datetime
import logging

from .accessor import Accessor
from .util import generate_message_control_id

logger = logging.getLogger(__file__)

_SENTINEL = object()


class Sequence(list):
    """Base class for sequences that can be indexed using 1-based index"""

    def __call__(self, index, value=_SENTINEL):
        """Support list access using HL7 compatible 1-based indices.
        Can be used to get and set values.

        >>> s = hl7.Sequence([1, 2, 3, 4])
        >>> s(1) == s[0]
        True
        >>> s(2, "new")
        >>> s
        [1, 'new', 3, 4]
        """
        index = self._adjust_index(int(index))
        if value is _SENTINEL:
            return self[index]
        else:
            self[index] = value

    def _adjust_index(self, index):
        """Subclasses can override if they do not want HL7 1-based indexing when used as callable"""
        if index >= 1:
            return index - 1
        else:
            return index


class Container(Sequence):
    """Abstract root class for the parts of the HL7 message."""

    def __init__(
        self, separator, sequence=[], esc="\\", separators="\r|~^&", factory=None
    ):
        # Initialize the list object, optionally passing in the
        # sequence.  Since list([]) == [], using the default
        # parameter will not cause any issues.
        super(Container, self).__init__(sequence)
        self.separator = separator
        self.esc = esc
        self.separators = separators
        self.factory = factory if factory is not None else Factory

    def __getitem__(self, item):
        # Python slice operator was returning a regular list, not a
        # Container subclass
        sequence = super(Container, self).__getitem__(item)
        if isinstance(item, slice):
            return self.__class__(
                self.separator,
                sequence,
                self.esc,
                self.separators,
                factory=self.factory,
            )
        return sequence

    def __getslice__(self, i, j):
        # Python 2.x compatibility.  __getslice__ is deprecated, and
        # we want to wrap the logic from __getitem__ when handling slices
        return self.__getitem__(slice(i, j))

    def __str__(self):
        """Join a the child containers into a single string, separated
        by the self.separator.  This method acts recursively, calling
        the children's __unicode__ method.  Thus ``unicode()`` is the
        approriate method for turning the python-hl7 representation of
        HL7 into a standard string.

        >>> str(h) == message
        True

        """
        return self.separator.join((str(x) for x in self))


class Message(Container):
    """Representation of an HL7 message. It contains a list
    of :py:class:`hl7.Segment` instances.
    """

    def __getitem__(self, key):
        """Index, segment-based or accessor lookup.

        If key is an integer, ``__getitem__`` acts list a list, returning
        the :py:class:`hl7.Segment` held at that index:

        >>> h[1]  # doctest: +ELLIPSIS
        [['PID'], ...]

        If the key is a string of length 3, ``__getitem__`` acts like a dictionary,
        returning all segments whose *segment_id* is *key*
        (alias of :py:meth:`hl7.Message.segments`).

        >>> h['OBX']  # doctest: +ELLIPSIS
        [[['OBX'], ['1'], ...]]

        If the key is a string of length greater than 3,
        the key is parsed into an :py:class:`hl7.Accessor` and passed
        to :py:meth:`hl7.Message.extract_field`.

        If the key is an :py:class:`hl7.Accessor`, it is passed to
        :py:meth:`hl7.Message.extract_field`.
        """
        if isinstance(key, str):
            if len(key) == 3:
                return self.segments(key)
            return self.extract_field(*Accessor.parse_key(key))
        elif isinstance(key, Accessor):
            return self.extract_field(*key)
        return super(Message, self).__getitem__(key)

    def __setitem__(self, key, value):
        """Index or accessor assignment.

        If key is an integer, ``__setitem__`` acts list a list, setting
        the :py:class:`hl7.Segment` held at that index:

        >>> h[1] = hl7.Segment("|", [hl7.Field("^", ['PID'], [''])])

        If the key is a string of length greater than 3,
        the key is parsed into an :py:class:`hl7.Accessor` and passed
        to :py:meth:`hl7.Message.assign_field`.

        >>> h["PID.2"] = "NEW"

        If the key is an :py:class:`hl7.Accessor`, it is passed to
        :py:meth:`hl7.Message.assign_field`.
        """
        if isinstance(key, str) and len(key) > 3 and isinstance(value, str):
            return self.assign_field(value, *Accessor.parse_key(key))
        elif isinstance(key, Accessor):
            return self.assign_field(value, *key)
        return super(Message, self).__setitem__(key, value)

    def segment(self, segment_id):
        """Gets the first segment with the *segment_id* from the parsed
        *message*.

        >>> h.segment('PID')  # doctest: +ELLIPSIS
        [['PID'], ...]

        :rtype: :py:class:`hl7.Segment`
        """
        # Get the list of all the segments and pull out the first one,
        # if possible
        match = self.segments(segment_id)
        # We should never get an IndexError, since segments will instead
        # throw an KeyError
        return match[0]

    def segments(self, segment_id):
        """Returns the requested segments from the parsed *message* that are
        identified by the *segment_id* (e.g. OBR, MSH, ORC, OBX).

        >>> h.segments('OBX')
        [[['OBX'], ['1'], ...]]

        :rtype: list of :py:class:`hl7.Segment`
        """
        # Compare segment_id to the very first string in each segment,
        # returning all segments that match.
        # Return as a Sequence so 1-based indexing can be used
        matches = Sequence(segment for segment in self if segment[0][0] == segment_id)
        if len(matches) == 0:
            raise KeyError("No %s segments" % segment_id)
        return matches

    def extract_field(
        self,
        segment,
        segment_num=1,
        field_num=1,
        repeat_num=1,
        component_num=1,
        subcomponent_num=1,
    ):
        """
            Extract a field using a future proofed approach, based on rules in:
            http://wiki.medical-objects.com.au/index.php/Hl7v2_parsing

            'PID|Field1|Component1^Component2|Component1^Sub-Component1&Sub-Component2^Component3|Repeat1~Repeat2',

                |   PID.F3.R1.C2.S2 = 'Sub-Component2'
                |   PID.F4.R2.C1 = 'Repeat1'

            Compatibility Rules:

                If the parse tree is deeper than the specified path continue
                following the first child branch until a leaf of the tree is
                encountered and return that value (which could be blank).

                Example:

                    |   PID.F3.R1.C2 = 'Sub-Component1' (assume .SC1)

                If the parse tree terminates before the full path is satisfied
                check each of the subsequent paths and if every one is specified
                at position 1 then the leaf value reached can be returned as the
                result.

                    |   PID.F4.R1.C1.SC1 = 'Repeat1'    (ignore .SC1)
        """
        # Save original values for error messages
        accessor = Accessor(
            segment, segment_num, field_num, repeat_num, component_num, subcomponent_num
        )

        field_num = field_num or 1
        repeat_num = repeat_num or 1
        component_num = component_num or 1
        subcomponent_num = subcomponent_num or 1

        segment = self.segments(segment)(segment_num)
        if field_num < len(segment):
            field = segment(field_num)
        else:
            if repeat_num == 1 and component_num == 1 and subcomponent_num == 1:
                return ""  # Assume non-present optional value
            raise IndexError("Field not present: {0}".format(accessor.key))

        rep = field(repeat_num)

        if not isinstance(rep, Repetition):
            # leaf
            if component_num == 1 and subcomponent_num == 1:
                return (
                    rep
                    if accessor.segment == "MSH" and accessor.field_num in (1, 2)
                    else self.unescape(rep)
                )
            raise IndexError(
                "Field reaches leaf node before completing path: {0}".format(
                    accessor.key
                )
            )

        if component_num > len(rep):
            if subcomponent_num == 1:
                return ""  # Assume non-present optional value
            raise IndexError("Component not present: {0}".format(accessor.key))

        component = rep(component_num)
        if not isinstance(component, Component):
            # leaf
            if subcomponent_num == 1:
                return self.unescape(component)
            raise IndexError(
                "Field reaches leaf node before completing path: {0}".format(
                    accessor.key
                )
            )

        if subcomponent_num <= len(component):
            subcomponent = component(subcomponent_num)
            return self.unescape(subcomponent)
        else:
            return ""  # Assume non-present optional value

    def assign_field(
        self,
        value,
        segment,
        segment_num=1,
        field_num=None,
        repeat_num=None,
        component_num=None,
        subcomponent_num=None,
    ):
        """
            Assign a value into a message using the tree based assignment notation.
            The segment must exist.

            Extract a field using a future proofed approach, based on rules in:
            http://wiki.medical-objects.com.au/index.php/Hl7v2_parsing
        """
        segment = self.segments(segment)(segment_num)

        while len(segment) <= field_num:
            segment.append(self.create_field([]))
        field = segment(field_num)
        if repeat_num is None:
            field[:] = [value]
            return
        while len(field) < repeat_num:
            field.append(self.create_repetition([]))
        repetition = field(repeat_num)
        if component_num is None:
            repetition[:] = [value]
            return
        while len(repetition) < component_num:
            repetition.append(self.create_component([]))
        component = repetition(component_num)
        if subcomponent_num is None:
            component[:] = [value]
            return
        while len(component) < subcomponent_num:
            component.append("")
        component(subcomponent_num, value)

    def escape(self, field, app_map=None):
        """
            See: http://www.hl7standards.com/blog/2006/11/02/hl7-escape-sequences/

            To process this correctly, the full set of separators (MSH.1/MSH.2) needs to be known.

            Pass through the message. Replace recognised characters with their escaped
            version. Return an ascii encoded string.

            Functionality:

            *   Replace separator characters (2.10.4)
            *   replace application defined characters (2.10.7)
            *   Replace non-ascii values with hex versions using HL7 conventions.

            Incomplete:

            *   replace highlight characters (2.10.3)
            *   How to handle the rich text substitutions.
            *   Merge contiguous hex values
        """
        if not field:
            return field

        esc = str(self.esc)

        DEFAULT_MAP = {
            self.separators[1]: "F",  # 2.10.4
            self.separators[2]: "R",
            self.separators[3]: "S",
            self.separators[4]: "T",
            self.esc: "E",
            "\r": ".br",  # 2.10.6
        }

        rv = []
        for offset, c in enumerate(field):
            if app_map and c in app_map:
                rv.append(esc + app_map[c] + esc)
            elif c in DEFAULT_MAP:
                rv.append(esc + DEFAULT_MAP[c] + esc)
            elif ord(c) >= 0x20 and ord(c) <= 0x7E:
                rv.append(c)
            else:
                rv.append("%sX%2x%s" % (esc, ord(c), esc))

        return "".join(rv)

    def unescape(self, field, app_map=None):  # noqa: C901
        """
            See: http://www.hl7standards.com/blog/2006/11/02/hl7-escape-sequences/

            To process this correctly, the full set of separators (MSH.1/MSH.2) needs to be known.

            This will convert the identifiable sequences.
            If the application provides mapping, these are also used.
            Items which cannot be mapped are removed

            For example, the App Map count provide N, H, Zxxx values

            Chapter 2: Section 2.10

            At the moment, this functionality can:

            *   replace the parsing characters (2.10.4)
            *   replace highlight characters (2.10.3)
            *   replace hex characters. (2.10.5)
            *   replace rich text characters (2.10.6)
            *   replace application defined characters (2.10.7)

            It cannot:

            *   switch code pages / ISO IR character sets
        """
        if not field or field.find(self.esc) == -1:
            return field

        DEFAULT_MAP = {
            "H": "_",  # Override using the APP MAP: 2.10.3
            "N": "_",  # Override using the APP MAP
            "F": self.separators[1],  # 2.10.4
            "R": self.separators[2],
            "S": self.separators[3],
            "T": self.separators[4],
            "E": self.esc,
            ".br": "\r",  # 2.10.6
            ".sp": "\r",
            ".fi": "",
            ".nf": "",
            ".in": "    ",
            ".ti": "    ",
            ".sk": " ",
            ".ce": "\r",
        }

        rv = []
        collecting = []
        in_seq = False
        for offset, c in enumerate(field):
            if in_seq:
                if c == self.esc:
                    in_seq = False
                    value = "".join(collecting)
                    collecting = []
                    if not value:
                        logger.warn(
                            "Error unescaping value [%s], empty sequence found at %d",
                            field,
                            offset,
                        )
                        continue
                    if app_map and value in app_map:
                        rv.append(app_map[value])
                    elif value in DEFAULT_MAP:
                        rv.append(DEFAULT_MAP[value])
                    elif value.startswith(".") and (
                        (app_map and value[:3] in app_map) or value[:3] in DEFAULT_MAP
                    ):
                        # Substitution with a number of repetitions defined (2.10.6)
                        if app_map and value[:3] in app_map:
                            ch = app_map[value[:3]]
                        else:
                            ch = DEFAULT_MAP[value[:3]]
                        count = int(value[3:])
                        rv.append(ch * count)

                    elif (
                        value[0] == "C"
                    ):  # Convert to new Single Byte character set : 2.10.2
                        # Two HEX values, first value chooses the character set (ISO-IR), second gives the value
                        logger.warn(
                            "Error inline character sets [%s] not implemented, field [%s], offset [%s]",
                            value,
                            field,
                            offset,
                        )
                    elif (
                        value[0] == "M"
                    ):  # Switch to new Multi Byte character set : 2.10.2
                        # Three HEX values, first value chooses the character set (ISO-IR), rest give the value
                        logger.warn(
                            "Error inline character sets [%s] not implemented, field [%s], offset [%s]",
                            value,
                            field,
                            offset,
                        )
                    elif value[0] == "X":  # Hex encoded Bytes: 2.10.5
                        value = value[1:]
                        try:
                            for off in range(0, len(value), 2):
                                rv.append(chr(int(value[off : off + 2], 16)))
                        except Exception:
                            logger.exception(
                                "Error decoding hex value [%s], field [%s], offset [%s]",
                                value,
                                field,
                                offset,
                            )
                    else:
                        logger.exception(
                            "Error decoding value [%s], field [%s], offset [%s]",
                            value,
                            field,
                            offset,
                        )
                else:
                    collecting.append(c)
            elif c == self.esc:
                in_seq = True
            else:
                rv.append(str(c))

        return "".join(rv)

    def create_message(self, seq):
        """Create a new :py:class:`hl7.Message` compatible with this message"""
        return self.factory.create_message(
            self.separators[0],
            seq,
            esc=self.esc,
            separators=self.separators,
            factory=self.factory,
        )

    def create_segment(self, seq):
        """Create a new :py:class:`hl7.Segment` compatible with this message"""
        return self.factory.create_segment(
            self.separators[1],
            seq,
            esc=self.esc,
            separators=self.separators[1:],
            factory=self.factory,
        )

    def create_field(self, seq):
        """Create a new :py:class:`hl7.Field` compatible with this message"""
        return self.factory.create_field(
            self.separators[2],
            seq,
            esc=self.esc,
            separators=self.separators[2:],
            factory=self.factory,
        )

    def create_repetition(self, seq):
        """Create a new :py:class:`hl7.Repetition` compatible with this message"""
        return self.factory.create_repetition(
            self.separators[3],
            seq,
            esc=self.esc,
            separators=self.separators[3:],
            factory=self.factory,
        )

    def create_component(self, seq):
        """Create a new :py:class:`hl7.Component` compatible with this message"""
        return self.factory.create_component(
            self.separators[4],
            seq,
            esc=self.esc,
            separators=self.separators[4:],
            factory=self.factory,
        )

    def create_ack(
        self, ack_code="AA", message_id=None, application=None, facility=None
    ):
        """
        Create an hl7 ACK response :py:class:`hl7.Message`, per spec 2.9.2, for this message.

        See http://www.hl7standards.com/blog/2007/02/01/ack-message-original-mode-acknowledgement/

        ``ack_code`` options are one of `AA` (Application Accept), `AR` (Application Reject),
        `AE` (Application Error), `CA` (Commit Accept - Enhanced Mode),
        `CR` (Commit Reject - Enhanced Mode), or `CE` (Commit Error - Enhanced Mode)
        (see HL7 Table 0008 - Acknowledgment Code)
        ``message_id`` control message ID for ACK, defaults to unique generated ID
        ``application`` name of sending application, defaults to receiving application of message
        ``facility`` name of sending facility, defaults to receiving facility of message
        """
        source_msh = self.segment("MSH")
        msh = self.create_segment([self.create_field(["MSH"])])
        msa = self.create_segment([self.create_field(["MSA"])])
        ack = self.create_message([msh, msa])

        ack.assign_field(str(source_msh(1)), "MSH", 1, 1)
        ack.assign_field(str(source_msh(2)), "MSH", 1, 2)
        # Sending application is source receving application
        ack.assign_field(
            str(application) if application is not None else str(source_msh(5)),
            "MSH",
            1,
            3,
        )
        # Sending facility is source receving facility
        ack.assign_field(
            str(facility) if facility is not None else str(source_msh(6)), "MSH", 1, 4
        )
        # Receiving application is source sending application
        ack.assign_field(str(source_msh(3)), "MSH", 1, 5)
        # Receiving facility is source sending facility
        ack.assign_field(str(source_msh(4)), "MSH", 1, 6)
        ack.assign_field(
            str(datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S")), "MSH", 1, 7
        )
        # Message type code
        ack.assign_field("ACK", "MSH", 1, 9, 1, 1)
        # Copy trigger event from source
        ack.assign_field(str(source_msh(9)(1)(2)), "MSH", 1, 9, 1, 2)
        ack.assign_field("ACK", "MSH", 1, 9, 1, 3)
        ack.assign_field(
            message_id if message_id is not None else generate_message_control_id(),
            "MSH",
            1,
            10,
        )
        ack.assign_field(str(source_msh(11)), "MSH", 1, 11)
        ack.assign_field(str(source_msh(12)), "MSH", 1, 12)

        ack.assign_field(str(ack_code), "MSA", 1, 1)
        ack.assign_field(str(source_msh(10)), "MSA", 1, 2)

        return ack

    def __str__(self):
        # Per spec, Message Construction Rules, Section 2.6 (v2.8), Message ends
        # with the carriage return
        return super(Message, self).__str__() + self.separator


class Segment(Container):
    """Second level of an HL7 message, which represents an HL7 Segment.
    Traditionally this is a line of a message that ends with a carriage
    return and is separated by pipes. It contains a list of
    :py:class:`hl7.Field` instances.
    """

    def _adjust_index(self, index):
        # First element is the segment name, so we don't need to adjust to get 1-based
        return index

    def __str__(self):
        if str(self[0]) in ["MSH", "FHS"]:
            return (
                str(self[0])
                + str(self[1])
                + str(self[2])
                + str(self[1])
                + self.separator.join((str(x) for x in self[3:]))
            )
        return super(Segment, self).__str__()


class Field(Container):
    """Third level of an HL7 message, that traditionally is surrounded
    by pipes and separated by carets. It contains a list of strings
    or :py:class:`hl7.Repetition` instances.
    """


class Repetition(Container):
    """Fourth level of an HL7 message. A field can repeat.
    It contains a list of strings or :py:class:`hl7.Component` instances.
    """


class Component(Container):
    """Fifth level of an HL7 message. A component is a composite datatypes.
    It contains a list of string sub-components.
    """


class Factory(object):
    """Factory used to create each type of Container.

    A subclass can be used to create specialized subclasses of each container.
    """

    create_message = Message  #: Create an instance of :py:class:`hl7.Message`
    create_segment = Segment  #: Create an instance of :py:class:`hl7.Segment`
    create_field = Field  #: Create an instance of :py:class:`hl7.Field`
    create_repetition = Repetition  #: Create an instance of :py:class:`hl7.Repetition`
    create_component = Component  #: Create an instance of :py:class:`hl7.Component`
